# non_continuous_timelapse.py
# Copyright (C) 2011 Julien Miotte <miotte.julien@gmail.com>
#
# This module is part of gfbi_core and is released under the GPLv3
# License: http://www.gnu.org/licenses/gpl-3.0.txt

from datetime import time, timedelta, datetime
from gfbi_core.util import GfbiException

DEFAULT_AUTHORIZED_HOURS = ((time.min, time.max),)
DEFAULT_AUTHORIZED_WEEKDAYS = (0, 1, 2, 3, 4, 5, 6)

class non_continuous_timelapse:
    def __init__(self, authorized_dates,
                 authorized_hours=DEFAULT_AUTHORIZED_HOURS,
                 authorized_weekdays=DEFAULT_AUTHORIZED_WEEKDAYS):
        """
            Simulates a continuous timelapse out of hours, dates and weekdays
            limits.

            :param min_date:
                datetime object describing the min authorized date
            :param max_date:
                datetime object describing the max authorized date
            :param authorized_hours:
                tuple containing 2-tuples of the limits of the authorized time
                ranges
            :param authorized_weekdays:
                tuple containing the authorized weekdays, described by their
                number in a week starting by monday -> 1.
        """
        self.authorized_ranges = {}
        self.total_days = 0
        self.total_seconds = 0

        min_date, max_date = authorized_dates
        days_lapse = (max_date - min_date).days

        cur_date = min_date

        while cur_date != max_date:
            if cur_date.weekday() in authorized_weekdays:
                self.total_days += 1
                for time_min, time_max in authorized_hours:
                    down_limit = datetime(
                        cur_date.year, cur_date.month, cur_date.day,
                        time_min.hour, time_min.minute, time_min.second,
                        time_min.microsecond)
                    up_limit = datetime(
                        cur_date.year, cur_date.month, cur_date.day,
                        time_max.hour, time_max.minute, time_max.second,
                        time_max.microsecond)

                    delta = (up_limit - down_limit)
                    self.authorized_ranges[self.total_seconds] = (down_limit,
                                                                  up_limit)

                    self.total_seconds += delta.seconds

            cur_date += timedelta(1)

        if not self.authorized_ranges:
            raise GfbiException("The non-continuous timelapse is empty.")

    def get_total_seconds(self):
        """
            Returns the number of seconds of the simulated timelapse.
        """
        return self.total_seconds

    def datetime_from_seconds(self, seconds):
        """
            Returns an absolute datetime out of a relative number of seconds
            since the beggining of the simulated timelapse.

            :param seconds:
                The relative number of seconds since the beggining of the
                simulated timelapse.
        """
        keys = sorted(self.authorized_ranges)
        keys.reverse()

        stamp = 0
        for stamp in keys:
            if seconds > stamp:
                break

        min_date, max_date = self.authorized_ranges[stamp]

        delta_seconds = seconds - stamp
        return min_date + timedelta(0, delta_seconds)





# git_model.py
# Copyright (C) 2011 Julien Miotte <miotte.julien@gmail.com>
#
# This module is part of gfbi_core and is released under the GPLv3
# License: http://www.gnu.org/licenses/gpl-3.0.txt

import sys
from git import Repo
from git.objects.util import altz_to_utctz_str

from gfbi_core.util import Timezone, DummyCommit, DummyBranch, GfbiException, \
                           Index
from gfbi_core import ACTOR_FIELDS, TIME_FIELDS


class GitModel:
    """
        This class represents the list of commits of the current branch of a
        given repository. This class is meant to be Qt-free so that it can be
        used in other ways than gitbuster.
    """

    def __init__(self, directory=".", fake_branch_name="", from_commits=False,
                 remote_ref=None):
        """
            Initializes the model with the repository root directory.

            :param directory:
                Root directory of the git repository.
        """
        self._directory = directory

        self._remote_ref = False
        self._current_branch = None

        if fake_branch_name:
            # This is an empy gitModel that will be filled with data from
            # another model
            self._repo = None
            self._current_branch = DummyBranch(fake_branch_name)

            if not from_commits:
                raise GfbiException("Can't build a fake model without commits")

            self._from_commits = from_commits
        elif remote_ref:
            # This is a model on a remote repository
            self._repo = Repo(directory)
            self._remote_ref = remote_ref
            self._current_branch = False
        else:
            self._repo = Repo(directory)
            self._current_branch = self._repo.active_branch

        self._columns = ['hexsha',
                         'authored_date', 'committed_date',
                         'author_name', 'author_email',
                         'committer_name', 'committer_email',
                         'message', 'parents', 'tree', 'children']

        self._changed_branch_once = False
        self._commits = []
        self._unpushed = []
        self._children = {}

        self._old_branch_name = ""

    def is_fake_model(self):
        return isinstance(self._current_branch, DummyBranch)

    def is_remote_model(self):
        """
            Returns True if the model is build with a remote.
        """
        return self._remote_info

    def populate(self):
        """
            Populates the model, by constructing a list of the commits of the
            current branch of the given repository.
        """
        if self.is_fake_model():
            if not self._from_commits:
                error = "This is a fake model, but we don't have commits to build it."
                raise GfbiException(error)
            self._commits = list(self._from_commits)
            return

        self._commits = []
        self._unpushed = []
        self._children = {}

        if self._remote_ref:
            branch_rev = self._remote_ref.commit
        else:
            branch_rev = self._current_branch.commit

        if self._current_branch and self._current_branch.tracking_branch():
            remote_commits_head = self._current_branch.tracking_branch().commit
        else:
            remote_commits_head = None

        pushed = False
        for commit in self._repo.iter_commits(rev=branch_rev):
            self._commits.append(commit)
            for parent in commit.parents:
                if parent not in self._children:
                    self._children[parent] = [commit,]
                else:
                    self._children[parent].append(commit)

            if remote_commits_head is not None and \
               commit.hexsha == remote_commits_head.hexsha:
                pushed = True
            if not pushed:
                self._unpushed.append(commit)

    def is_commit_pushed(self, commit):
        """
            Returns True if the commit has been pushed to the remote branch.

            :param commit:
        """
        if isinstance(commit, DummyCommit):
            return False
        else:
            return not commit in self._unpushed

    def get_branches(self):
        """
            Returns the repository avalaible branches.
        """
        if self._repo:
            return self._repo.branches
        else:
            raise Exception("There is no branches here")

    def get_current_branch(self):
        """
            Returns the model's current branch (maybe a DummyBranch).
        """
        return self._current_branch

    def get_remote_ref(self):
        """
            Returns the model's remote reference.
        """
        return self._remote_ref

    def set_current_branch(self, branch, force=False):
        """
            Sets the model's current branch.
            As populate() is costly, we don't do it automaticaly here.

            :param branch:
                The desired branch to modelize.
            :param force:
                By default, users shouldn't change the branch of a model twice.
                They should create a new model.
        """
        if self._changed_branch_once and not force:
            raise GfbiException("You shouldn't change the branch twice.")

        if self.is_fake_model():
            # This is the moment after we wrote the model, the model is getting
            # real (not fake).
            self._repo = Repo(self._directory)

        self._current_branch = branch
        self._changed_branch_once = True
        self._old_branch_name = branch.name

    def get_commits(self):
        """
            Returns the commit list.
        """
        return self._commits

    def get_columns(self):
        """
            Returns the selected fields names.
        """
        return self._columns

    def get_column(self, field):
        """
            Since we're using this operation quite a lot, factorizing it.
        """
        return self._columns.index(field)

    def row_count(self):
        """
            Returns the count of commits.
        """
        return len(self._commits)

    def column_count(self):
        """
            Returns the count of selected fields names.
        """
        return len(self._columns)

    def data(self, index):
        """
            This method uses the index row to select the commit and the index
            column to select the field to return.

            :param index:
                The index of the wanted information.

            :return:
                Depending on the index column, one of the commit fields.
        """
        return self.orig_data(index)

    def orig_data(self, index):
        commit = self._commits[index.row()]
        column = index.column()
        field = self._columns[column]

        if field in TIME_FIELDS:
            if field == 'authored_date':
                _timestamp = commit.authored_date
                _utc_offset = altz_to_utctz_str(commit.author_tz_offset)
                _tz = Timezone(_utc_offset)
            elif field == 'committed_date':
                _timestamp = commit.committed_date
                _utc_offset = altz_to_utctz_str(commit.committer_tz_offset)
                _tz = Timezone(_utc_offset)
            value = (_timestamp, _tz)
        elif field in ACTOR_FIELDS:
            actor = eval("commit." + field.split('_')[0])
            if '_email' in field:
                value = actor.email
            else:
                value = actor.name
        elif field == "message":
            value = commit.message.rstrip()
        elif field == "children":
            if commit in self._children:
                value = list(self._children[commit])
            else:
                value = []
        elif field == "parents":
            value = list(commit.parents)
        else:
            value = eval("commit." + field)

        return value

    def row_of(self, commit):
        return self._commits.index(commit)

    def get_old_branch_name(self):
        """
            Returns the old name of the branch.
        """
        return self._old_branch_name

    def is_first_commit(self, index):
        """
            Returns true if the index points to the first commit.
            First commit is the last of _commits.
        """
        return index.row() == len(self._commits) - 1



# editable_git_model.py
# Copyright (C) 2011 Julien Miotte <miotte.julien@gmail.com>
#
# This module is part of gfbi_core and is released under the GPLv3
# License: http://www.gnu.org/licenses/gpl-3.0.txt

from time import mktime

from random import random
#from random import uniform

from gfbi_core import NAMES
from gfbi_core.util import DummyCommit, InsertAction, SetAction, RemoveAction, \
                           SetBranchNameAction, DummyBranch, GfbiException, \
                           Index
from gfbi_core.git_model import GitModel
from gfbi_core import TIME_FIELDS
from gfbi_core.git_filter_rebase import git_filter_rebase
from gfbi_core.non_continuous_timelapse import non_continuous_timelapse
from gfbi_core.validation import validate_branch_name


class EditableGitModel(GitModel):
    """
        This class represents the list of commits of the current branch of a
        given repository. This class is meant to be Qt-free so that it can be
        used in other ways than gitbuster.
    """

    def __init__(self, directory=".", fake_branch_name="", from_commits=False):
        """
            Initializes the model with the repository root directory.

            :param directory:
                Root directory of the git repository.
        """
        if fake_branch_name:
            # This is an empy gitModel that will be filled with data from
            # another model
            self.orig_model = None
        else:
            self.orig_model = GitModel(directory=directory)

        self.init_attributes()

        GitModel.__init__(self, directory=directory,
                          fake_branch_name=fake_branch_name,
                          from_commits=from_commits)

    def init_attributes(self):
        """
            These attributes should be resetted when populating the model.
        """
        self._modifications = {}
        self._deleted_commits = []
        self._merge = False
        self._history = []
        self._last_history_event = -1
        self._conflicting_commit = None
        self._unmerged_files = None
        self._solutions = {}
        self._new_branch_name = ""
        self._git_process = None
        self._start_write_cache = {}
        self._children_cache = {}

    def populate(self):
        """
            Populates the model, by constructing a list of the commits of the
            current branch of the given repository.
        """
        self.init_attributes()

        if not self.is_fake_model():
            self.orig_model.populate()

        GitModel.populate(self)

    def set_current_branch(self, branch, force=False):
        """
            Sets the model's current branch.

            :param branch:
                The desired branch to modelize.
        """
        if self._changed_branch_once and not force:
            raise GfbiException("You shouldn't change the branch twice.")

        if self.is_fake_model():
            # This is the moment after we wrote the model, the model is getting
            # real (not fake).
            self.orig_model = GitModel(directory=self._directory)

        self.orig_model.set_current_branch(branch, force=force)
        self._modifications = {}

        GitModel.set_current_branch(self, branch, force=force)

    def get_modifications(self):
        """
            Returns the modified values dictionnary.
        """
        return self._modifications

    def get_modified_count(self):
        """
            Returns the number of modified or deleted commits.
        """
        modified_or_deleted = set(
            [commit for commit in self._commits
             if self.commit_is_modified(commit) and
                not self.is_deleted(commit)])
        return len(modified_or_deleted)

    def get_orig_model(self):
        """
            Returns an unmodified GitModel.
        """
        return self.orig_model

    def data(self, index):
        """
            This method uses the index row to select the commit and the index
            column to select the field to return.

            :param index:
                The index of the wanted information.

            :return:
                Depending on the index column, one of the commit fields.
        """
        if self.is_modified(index):
            return self.modified_data(index)
        else:
            return self.orig_data(index)

    def c_data(self, commit, field):
        """
            This is a convenient method to access data using the commit and
            the column.
        """
        row = self.row_of(commit)
        col = self.get_column(field)
        return self.data(Index(row, col))

    def modified_data(self, index):
        commit = self._commits[index.row()]
        column = index.column()
        field = self._columns[column]

        value = ""
        if commit in self._modifications:
            modification = self._modifications[commit]
            value = modification[field]

            if value and field in ("children", "parents"):
                value = list(value)

        return value

    def set_merge(self, merge_state):
        """
            Set the merge option, meaning that the committed and authored
            notions are merged. For instance, if the committed date is changed
            in the set_data() method, the authored date will be set with the
            same value.

            :param merge_state:
                Boolean, sets the merge option.
        """
        self._merge = merge_state

    def set_data(self, index, value, ignore_history=False):
        """
            Set the given value to the commit and the field determined by the
            index.

            :param index:
                The index of the commit and the field that should be modified.
            :param value:
                The value that will be assigned.
        """
        if not 0 <= index.row() < len(self._commits):
            raise GfbiException("Invalid index")

        commit = self._commits[index.row()]
        column = index.column()
        field_name = self._columns[column]

        reference = None

        if field_name in TIME_FIELDS:
            if self.data(index) is not None:
                reference, tz = self.data(index)
        else:
            reference = self.data(index)

        # This is useless in the development version of GitPython
        # See https://github.com/gitpython-developers/GitPython/commit/096897123ab5d8b500024e63ca81b658f3cb93da
        git_python_precheck = (hasattr(reference, 'binsha') !=
                               hasattr(value, 'binsha'))

        if git_python_precheck or reference != value:
            self.set_field_data(commit, field_name, value)

            if not ignore_history:
                action = SetAction(index, reference, value)
                self._history[self._last_history_event].append(action)

            if self._merge:
                if field_name == "committed_date":
                    self.set_field_data(commit, "authored_date", value)
                elif field_name == "authored_date":
                    self.set_field_data(commit, "committed_date", value)
                elif field_name == "author_name":
                    self.set_field_data(commit, "committer_name", value)
                elif field_name == "committer_name":
                    self.set_field_data(commit, "author_name", value)
                elif field_name == "author_email":
                    self.set_field_data(commit, "committer_email", value)
                elif field_name == "committer_email":
                    self.set_field_data(commit, "author_email", value)

    def set_field_data(self, commit, field, value):
        """
            This method is used in set_data() to assign the given value to the
            given field of the given commit.

            :param commit:
                The commit that will be modified.
            :param field:
                The field that will be modified.
            :param value:
                The value that will be assigned.
        """
        if commit not in self._modifications:
            self._modifications[commit] = {}
        self._modifications[commit][field] = value

    def start_history_event(self):
        """
            Start a new history event. If the current event isn't the last one,
            drop every event after the current event.
        """
        while self._last_history_event < len(self._history) - 1:
            self._history.pop()

        self._last_history_event += 1
        self._history.append([])

    def undo_history(self):
        """
            Reverts the history one event back.
        """
        if self._last_history_event >= 0:
            for action in reversed(self._history[self._last_history_event]):
                action.undo(self)
            if self._last_history_event > -1:
                self._last_history_event -= 1

    def redo_history(self):
        """
            Replays the history one event forward.
        """
        if self._last_history_event < len(self._history) - 1:
            self._last_history_event += 1
            for action in self._history[self._last_history_event]:
                action.redo(self)

    def undelete_commit(self, commit, modifications):
        """
            Remove a commit from the _deleted_commits list.
        """
        self._deleted_commits.remove(commit)

        if modifications:
            self._modifications[commit] = modifications

    def insert_commit(self, row, commit, modifications):
        """
            The parent commit is the previous commit in the history.
        """
        self._commits.insert(row, commit)

        if modifications:
            self._modifications[commit] = modifications

    def insert_rows(self, position, rows):
        """
            Inserts blank rows in the model.

            :param position:
                Position from where to insert.
            :param rows:
                Number of rows to insert.
        """
        for i in xrange(rows):
            commit = DummyCommit()
            self._commits.insert(position, commit)

            self._modifications[commit] = {}
            for field in NAMES:
                self._modifications[commit][field] = None

            action = InsertAction(position, commit, self._modifications[commit])
            self._history[self._last_history_event].append(action)

    def remove_rows(self, position, rows, ignore_history=False,
                    really_remove=False):
        """
            Removes rows from the model.

            :param position:
                Position from where to delete.
            :param rows:
                Number of rows to delete.
        """
        for i in xrange(rows):
            if really_remove:
                commit = self._commits.pop(position)
                self._modifications.pop(commit)
            else:
                commit = self._commits[position + i]
                if not self.is_deleted(commit):
                    self._deleted_commits.append(commit)

                    if not ignore_history:
                        modifications = None
                        if self._modifications.has_key(commit):
                            modifications = self._modifications[commit]
                        action = RemoveAction(position, commit, modifications)
                        self._history[self._last_history_event].append(action)

    def is_deleted(self, indexorcommit):
        """
            If indexorcommit:
                * is an index, if the commit at index.row() is a deleted commit,
                return True.
                * is a commit, if the commit is a deleted commit, return True.
        """
        if hasattr(indexorcommit, 'row'):
            if not 0 <= indexorcommit.row() < len(self._commits):
                raise GfbiException("Invalid index")
            commit = self._commits[indexorcommit.row()]
        else:
            commit = indexorcommit
        return commit in self._deleted_commits

    def is_inserted_commit(self, index):
        """
            If the commit at index.row() is a DummyCommit, return True.
        """
        commit = self._commits[index.row()]
        return isinstance(commit, DummyCommit)

    def is_modified(self, index):
        """
            Returns True if the commit field determined by the index has been
            modified (if there is a corresponding entry in the _modifications
            dict).

            :param index:
                Index of the field of the commit.

            :return:
                True if the field of the commit is modified else False.
        """
        commit = self._commits[index.row()]
        column = index.column()
        field_name = self._columns[column]

        mods = self._modifications

        if isinstance(commit, DummyCommit):
            return True

        if commit in mods and field_name in mods[commit]:
            return mods[commit][field_name] != self.orig_data(index)
        return False

    def commit_is_modified(self, commit):
        """
            Returns True is one of the commit fields has been modified.
        """
        row = self.row_of(commit)

        children_col = self.get_column("children")
        ignore_columns = set((children_col,))
        for col in set(xrange(self.column_count())) - ignore_columns:
            index = Index(row, col)
            if self.is_modified(index) and not self.is_deleted(index):
                return True
        return False

    def write(self, log=True, force_committed_date=False, dont_populate=False):
        """
            Start the git filter-branch command and therefore write the
            modifications stored in _modifications.

            :param log:
                Boolean, set to True to log the git command.
            :param force_committed_date:
                As the git way updates the committed author/date when
                cherry-picking, and since we offer to modify these values, we
                offer the user the choice to force the committed author/date
                or to let git update it.
        """
        self._git_process = git_filter_rebase(self, log=log,
                                    force_committed_date=force_committed_date,
                                    dont_populate=dont_populate)
                           # git_filter_branch_process(self,
                           #        directory=self._directory,
                           #        commits=self._commits,
                           #        modifications=self._modifications,
                           #        oldest_commit_parent=oldest_commit_parent,
                           #        log=log, script=script)

        self._git_process.start()

    def is_finished_writing(self):
        """
            Returns False if the git command process isn't finished else True.
        """
        if self._git_process is not None:
            return self._git_process.is_finished()
        return True

    def progress(self):
        """
            Returns the git command process progress.
        """
        if self._git_process is not None:
            return self._git_process.progress()
        return 0

    def is_write_success(self):
        """
            Returns True if the write process went through without failing.
        """
        if self._git_process is not None:
            return self._git_process.is_success()
        else:
            return False

    def get_start_write_from(self):
        """
            Returns the parents of the commits that are modified.
        """
        modified_commits = tuple(self._modifications.keys())
        if modified_commits in self._start_write_cache.keys():
            return self._start_write_cache[modified_commits]

        parents = set()
        for commit in self._modifications:
            if self.commit_is_modified(commit):
                parents.add(commit)

        smaller_parent_set = set(parents)
        for parent in parents:
            # Check that parent isn't in any of the other parents history tree
            if parent not in smaller_parent_set:
                continue

            for _parent in parents - set([parent]):
                # Go down the history tree of _parent
                this_parent_parents = self.c_data(_parent, "parents")
                if _parent in smaller_parent_set and \
                   parent in self.all_parents(_parent):
                    # Parent is present in the _parent history tree
                    # That means _parent must be removed from
                    # smaller_parent_set.
                    smaller_parent_set.remove(_parent)

        if parents == set([]) and self.is_fake_model():
            # Special case: no commit has really been modified, and this is a
            # fake model. We just need to update the top commit.
            smaller_parent_set = ([self._commits[0],])

        self._start_write_cache[modified_commits] = smaller_parent_set

        return smaller_parent_set

    def all_parents(self, commit):
        """
            Returns all the parents of a commit.
        """
        parents = set()

        parents_to_look = set([commit,])
        while parents_to_look:
            commit = parents_to_look.pop()
            for parent in self.c_data(commit, "parents"):
                yield parent
                parents_to_look.add(parent)

    def all_children(self, commits):
        """
            Returns a set with all the children of the given commits.
        """
        children = set()

        modified_commits = tuple(self._modifications.keys())
        if modified_commits in self._children_cache.keys():
            return self._children_cache[modified_commits]

        children_to_look = set(commits)
        while children_to_look:
            commit = children_to_look.pop()
            for child in self.c_data(commit, "children"):
                if not child in children:
                    children_to_look.add(child)

                children.add(child)

        self._children_cache[modified_commits] = children

        return children

    def get_to_rewrite_count(self):
        """
            Returns the number of commits to will be rewritten. That means the
            number of commit between HEAD and the oldest modified commit.
        """
        start_from_commits = self.get_start_write_from()
        all_children = self.all_children(start_from_commits)

        return len(all_children) + len(start_from_commits)

    def erase_modifications(self):
        """
            Erase all modifications: set _modified to {}.
        """
        self._modifications = {}

    def reorder_commits(self, dates, times, weekdays):
        """
            This method reorders the commits given specified timelapses and
            weekdays.
        """
        timelapse = non_continuous_timelapse(dates, times, weekdays)

        ## Random method
        #delta = truc_truc.get_total_seconds() / (how_many_commits + 1)
        #max_error = delta / 2
        #
        #time_cursor = 0
        #for commit in xrange(how_many_commits):
        #    time_cursor += delta
        #    # The next lines sets the commit_time to time_cursor, plus or less
        #    # an error
        #    new_commit_time = time_cursor + int((random() * 2 - 1) *max_error)

        # Uniform method
        total_seconds = timelapse.get_total_seconds()
        distribution = [int(random() * total_seconds)
                        for commit in xrange(len(self._commits))]
        distribution.sort()

        index = 0
        for commit in self._commits:
            this_distribution = distribution[index]
            new_commit_time = timelapse.datetime_from_seconds(
                                                            this_distribution)
            self.set_field_data(commit, "authored_date",
                                int(mktime(new_commit_time.timetuple())))
            self.set_field_data(commit, "committed_date",
                                int(mktime(new_commit_time.timetuple())))

            index += 1

    def set_conflicting_commit(self, row):
        """
            Sets the conflicting commit accordingly to the given row.
        """
        self._conflicting_commit = self._commits[row]

    def get_conflicting_commit(self):
        """
            Gets the conflicting commit.
        """
        return self._conflicting_commit

    def get_conflicting_row(self):
        """
            Returns the conflicting commit row.
        """
        return self._commits.index(self._conflicting_commit)

    def is_conflicting_commit(self, row):
        """
            Returns true is the given row points to the conflicting commit.
        """
        if self._conflicting_commit is None:
            return False

        return self._conflicting_commit == self._commits[row]

    def set_unmerged_files(self, u_files):
        """
            Sets the unmerged files as a dictionnary of :
                dict[filepath] = {"unmerged_content"    : unmerged_content,
                                  "diff"                : diff,
                                  "orig_content"        : orig_content,
                                  "git_status"          : git_status}

                where git_status is one of:
                    DD : both deleted
                    AU : added by us
                    UD : deleted by them
                    UA : added by them
                    DU : deleted by us
                    AA : both added
                    UU : both modified

                orig_content is the content of the file at filepath before the
                merge conflict

                unmerged_content is the content of the file at filepath at the
                moment of the merge conflict

                and diff is the diff that should be applied by the conflicting
                commit.
        """
        self._unmerged_files = u_files

    def get_unmerged_files(self):
        """
            Returns the unmerged files after cherry-picking the conflicting
            commit (self._conflicting_commit).
        """
        return self._unmerged_files

    def set_conflict_solutions(self, solutions):
        """
            Sets the solutions for the current conflicting commit.

            :param solutions:
                This is a dictionnary like:
                    solutions[filepath] = (action, custom_content)

                where filepath is the path of the unmerged file

                action is the action that should be taken after the conflict.
                Here is an explanation of what should happen for every action:
                    delete      : git rm filepath
                    add         : git add filepath
                    add_custom  : echo $custom_content > filepath
                                  git add filepath
        """
        self._solutions[self._conflicting_commit] = solutions

    def get_conflict_solutions(self):
        """
            Returns the solutions for every possible conflict of the model.
        """
        return self._solutions

    def is_valid_branch_name(self, name):
        """
            This allows us to check if the name is valid before setting it.
        """
        try:
            validate_branch_name(name)
        except Exception, err:
            return False, err
        else:
            return True, None

    def set_new_branch_name(self, name, ignore_history=False):
        """
            Set the new name of the branch.
        """
        name = name.strip()

        if not ignore_history:
            action = SetBranchNameAction(self._new_branch_name, name)
            self._history[self._last_history_event].append(action)

        if name == self._old_branch_name:
            self._new_branch_name = ""
        elif ignore_history and name == "":
            self._new_branch_name = ""
        else:
            validate_branch_name(name)
            self._new_branch_name = name

        return self._new_branch_name

    def get_new_branch_name(self):
        """
            Returns the new name of the branch.
        """
        return self._new_branch_name

    def is_name_modified(self):
        """
            Returns True if there is a new branch name to be written.
        """
        return self._new_branch_name != ""



import warnings
import sys

# We need to configure warnings since it is used to warn about import failures
warnings.simplefilter("always")

try:
    import git
except ImportError:
    warnings.warn("""Couldn't import git. You might want to install
                  GitPython from: http://pypi.python.org/pypi/GitPython/""",
                  ImportWarning)
    sys.exit(1)

str_maj, str_min, str_rev = git.__version__.split(".")
_maj, _min, _rev = int(str_maj), int(str_min), int(str_rev)
if  _maj < 0 or (_maj == 0 and _min < 3) or \
    (_maj == 0 and _min == 3 and _rev < 1):
    warnings.warn("This project needs GitPython (>=0.3.1).", ImportWarning)
    sys.exit(1)


# Needed by gitbuster
NAMES = {'actor_name':'Actor', 'author_name':'Author',
         'author_email':'Author Email',
         'authored_date':'Authored Date', 'committed_date':'Committed Date',
         'committer_name':'Committer', 'committer_email':'Committer Email',
         'count':'Count', 'diff':'Diff',
         'diffs':'Diffs', 'find_all':'Find All', 'hexsha':'Id',
         'lazy_properties':'Lazy Properties',
         'list_from_string':'List From String', 'message':'Message',
         'parents':'Parents', 'repo':'Repo', 'stats':'Stats',
         'summary':'Summary', 'tree':'Tree', 'children':'Children'}

NOT_EDITABLE_FIELDS = ['hexsha',]

ENV_FIELDS = {'author_name'     : 'GIT_AUTHOR_NAME',
              'author_email'    : 'GIT_AUTHOR_EMAIL',
              'authored_date'   : 'GIT_AUTHOR_DATE',
              'committer_name'  : 'GIT_COMMITTER_NAME',
              'committer_email' : 'GIT_COMMITTER_EMAIL',
              'committed_date'  : 'GIT_COMMITTER_DATE' }

TEXT_FIELDS = ['message', 'summary']
ACTOR_FIELDS = ['author_name', 'committer_name', 'author_email', 'committer_email']
TIME_FIELDS = ['authored_date', 'committed_date']

__version__ = "0.5b2"



# util.py
# Copyright (C) 2011 Julien Miotte <miotte.julien@gmail.com>
#
# This module is part of gfbi_core and is released under the GPLv3
# License: http://www.gnu.org/licenses/gpl-3.0.txt

from datetime import timedelta, tzinfo
from subprocess import Popen, PIPE
from git import Repo
import codecs


STATUSES = (
    ("both deleted:", "DD"),
    ("added by us:", "AU"),
    ("deleted by them:", "UD"),
    ("added by them:", "UA"),
    ("deleted by us:", "DU"),
    ("both added:", "AA"),
    ("both modified:", "UU")
)
REBASE_PATCH_FILE = ".git/rebase-merge/patch"


class Index:
    """
        This mimics the qModelIndex, so that we can use it in the
        GitModel.data() method when populating the model.
    """

    def __init__(self, row=0, column=0):
        """
            Initialization of the Index object.

            :param row:
                Row number, integer.

            :param column:
                Column number, integer.
        """
        self._row = row
        self._column = column

    def row(self):
        """
            Returns the row number.
        """
        return self._row

    def column(self):
        """
            Returns the column number.
        """
        return self._column


class Timezone(tzinfo):
    """
        Timezone class used to preserve the timezone information when handling
        the commits information.
    """

    def __init__(self, tz_string):
        """
            Initialize the Timezone object with it's string representation.

            :param tz_string:
                Representation of the offset to UTC of the timezone. (i.e.
                +0100 for CET or -0400 for ECT)
        """
        self.tz_string = tz_string

    def utcoffset(self, dt):
        """
            Returns the offset to UTC using the string representation of the
            timezone.

            :return:
                Timedelta object representing the offset to UTC.
        """
        sign = 1 if self.tz_string[0] == '+' else -1
        hour = sign * int(self.tz_string[1:-2])
        minutes = sign * int(self.tz_string[2:])
        return timedelta(hours=hour, minutes=minutes)

    def tzname(self, dt):
        """
            Returns the offset to UTC string representation.

            :return:
                Offset to UTC string representation.
        """
        return self.tz_string

    def dst(self, dt):
        """
            Returns a timedelta object representing a whole number of minutes
            with magnitude less than one day.

            :return:
                timedelta(0)
        """
        return timedelta(0)


class DummyCommit:

    def __init__(self):
        self.author_tzinfo = None
        self.committer_tzinfo = None
        self.binsha = 0


class HistoryAction:

    def revert(self, model):
        pass


class InsertAction(HistoryAction):

    def __init__(self, insert_position, commit, modifications):
        self._insert_position = insert_position
        self._commit = commit
        self._modifications = modifications

    def undo(self, model):
        model.remove_rows(self._insert_position, 1, ignore_history=True,
                          really_remove=True)

    def redo(self, model):
        model.insert_commit(self._insert_position, self._commit,
                            self._modifications)


class SetAction(HistoryAction):

    def __init__(self, set_index, old_value, new_value):
        self._set_index = set_index
        self._old_value = old_value
        self._new_value = new_value

    def undo(self, model):
        model.set_data(self._set_index, self._old_value, ignore_history=True)

    def redo(self, model):
        model.set_data(self._set_index, self._new_value, ignore_history=True)


class RemoveAction(HistoryAction):

    def __init__(self, remove_position, commit, modifications):
        self._remove_position = remove_position
        self._commit = commit
        self._modifications = modifications

    def undo(self, model):
        model.undelete_commit(self._commit, self._modifications)

    def redo(self, model):
        model.remove_rows(self._remove_position, 1, ignore_history=True)


class SetBranchNameAction(HistoryAction):

    def __init__(self, old_name, new_name):
        self._old_name = old_name
        self._new_name = new_name

    def undo(self, model):
        model.set_new_branch_name(self._old_name, ignore_history=True)

    def redo(self, model):
        model.set_new_branch_name(self._new_name, ignore_history=True)


class DummyBranch:

    def __init__(self, name):
        self.name = name
        self.path = name


class GfbiException(Exception):
    pass


def run_command(command):
    process = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
    output, errors = process.communicate()

    return output.split('\n'), errors.split('\n')


def get_unmerged_files(conflicting_hexsha, orig_hexsha, directory):
    """
        Collect several information about the current unmerged state.

        :param from_hexsha, to_hexsha:
            These parameters are used to provide the diff that should have been
            applied by the conflicting commit.
    """
    u_files = {}

    # Fetch diffs
    provide_unmerged_status(u_files)
    provide_diffs(u_files, conflicting_hexsha)
    provide_unmerged_contents(u_files)
    provide_orig_contents(u_files, orig_hexsha, directory)

    return u_files


def provide_unmerged_status(u_files):
    """
        Reads the output of 'git status' to find the unmerged statuses.
    """
    command = "git --no-pager status"
    output, errors = run_command(command)
    for line in output:
        for status, short_status in STATUSES:
            if status in line:
                u_file = line.split(status)[1].strip()
                git_status = short_status
                u_files.setdefault(u_file, {})["git_status"] = git_status
                break


def provide_diffs(u_files, conflicting_hexsha):
    """
        Process the output of git diff rather than calling git diff on
        every file in the path.
    """
    command = "git --no-pager diff %s~ %s" % \
            (conflicting_hexsha, conflicting_hexsha)
    diff_output, errors = run_command(command)

    u_file = None
    diff = ""
    for line in diff_output:
        if line[:10] == 'diff --git':
            if u_file and diff:
                u_files.setdefault(u_file, {})["diff"] = diff
                diff = ""
            u_file = line.split('diff --git a/')[1].split(' ')[0]

            if u_file not in u_files:
                # This isn't an unmerged file, don't log the diff
                u_file = None

        if u_file:
            diff += line + "\n"

    # Write the last diff
    if u_file:
        u_files.setdefault(u_file, {})["diff"] = diff


def provide_unmerged_contents(u_files):
    """
        Fetches the unmerged contents of the files.
    """
    # Make a backup of the unmerged file.
    for u_file, file_info in u_files.items():
        short_status = file_info["git_status"]
        unmerged_content = ""
        if not short_status == "DD":
            handle = codecs.open(u_file, encoding='utf-8', mode='r')
            unmerged_content = handle.read()
            handle.close()
        u_files.setdefault(u_file, {})["unmerged_content"] = unmerged_content


def provide_orig_contents(u_files, orig_hexsha, directory):
    """
        This method fetches the content of the files before the merge.

        Possible optimization: use git.repo.tree.
    """
    repo = Repo(directory)
    commit = repo.commit(rev=orig_hexsha)
    tree = commit.tree

    # Use Commit.tree[file].data_stream.read()
    for u_file, file_info in u_files.items():
        git_status = file_info["git_status"]
        orig_content = ""
        if git_status not in ('UA', 'DU', 'DD'):
            blob = tree[u_file]
            orig_content = blob.data_stream.read()
        u_files.setdefault(u_file, {})["orig_content"] = orig_content


def apply_solutions(solutions):
    """
        This apply the given solutions to the repository.

        :param solutions:
            See EditableGitModel.set_conflict_solutions.
    """
    for filepath, action in solutions.items():
        if action[0] == "delete":
            command = 'git rm %s'
        elif action[0] == "add":
            command = 'git add %s'
        elif action[0] == "add_custom":
            custom_content = action[1]
            handle = codecs.open(filepath, encoding='utf-8', mode='w')
            handle.write(custom_content)
            handle.close()
            command = 'git add %s'

        run_command(command % filepath)



# git_filter_rebase.py
# Copyright (C) 2011 Julien Miotte <miotte.julien@gmail.com>
#
# This module is part of gfbi_core and is released under the GPLv3
# License: http://www.gnu.org/licenses/gpl-3.0.txt

from threading import Thread
import os
import time
import codecs
from git import Repo

from gfbi_core.util import Index, run_command, apply_solutions, \
                           get_unmerged_files, GfbiException
from gfbi_core import ENV_FIELDS, ACTOR_FIELDS, TIME_FIELDS


def add_assign(commit_settings, field, value):
    commit_settings += ENV_FIELDS[field] + "='%s'" % value + " "
    return commit_settings


class git_filter_rebase(Thread):
    """
        Thread meant to execute and follow the progress of the git command
        process.
    """

    def __init__(self, parent, log=True, force_committed_date=False,
                 dont_populate=False):
        """
            Initialization of the GitFilterRebase thread.

            :param parent:
                GitModel object, parent of this thread.
            :param log:
                If set to True, the git commands will be logged.
            :param force_committed_date:
                As the git way updates the committed author/date when
                cherry-picking, and since we offer to modify these values, we
                offer the user the choice to force the committed author/date
                or to let git update it.
            :param dont_populate:
                If set to True, this process won't repopulate the model after
                the rewrite.
        """
        Thread.__init__(self)

        self._start_commits = parent.get_start_write_from()

        self._log = log
        if log:
            self._logfile = ".gitbuster_" + time.strftime("%d-%m-%Y.%H-%M")

        self._force_committed_date = force_committed_date
        self._dont_populate = dont_populate
        self._model = parent
        self._directory = parent._directory
        self._branch = parent._current_branch

        self._to_rewrite_count = self._model.get_to_rewrite_count()
        self._solutions = self._model.get_conflict_solutions()

        self._u_files = {}
        self._output = []
        self._errors = []
        self._updated_refs = {}
        self._should_be_updated = []
        self._last_updated_sha = None
        self._progress = None
        self._finished = False
        self._success = False

        if self._model.is_fake_model():
            self._fallback_branch_name = Repo(self._directory).branches[0].name
        else:
            self._fallback_branch_name = self._branch.name

        if not self.model_is_applicable():
            raise GfbiException("Can't apply, the repository changed.")

    def model_is_applicable(self):
        """
            Here we are going to check if the tip commit of the given model is
            the same as the real one. This is important because applying a
            dated model (if the repository has new commits since we launched
            gfbi_core) will cause data losses.
        """
        if self._model.is_fake_model():
            # No need to check fake models
            return True

        a_repo = Repo(self._directory)
        current_tip = a_repo.branches[self._branch.name].commit
        current_tip.hexsha

        row = 0
        while self._model.is_inserted_commit(Index(row, 0)):
            row += 1
        index = Index(row, 0)
        model_tip_hexsha = self._model.data(index)
        return current_tip.hexsha == model_tip_hexsha

    def all_should_be_updated(self, updated_parents):
        """
            Returns all the commits that should be updated, if the given commit
            would be modified.
        """
        should_be_updated = set()
        for updated_parent in updated_parents:
            for commit in self._model.c_data(updated_parent, "children"):
                should_be_updated.add(commit)
                should_be_updated.update(self.all_should_be_updated((commit,)))
        return should_be_updated

    def log(self, message):
        if self._log:
            handle = codecs.open(self._logfile, encoding='utf-8', mode='a')
            log_stamp = unicode(time.strftime("[%d-%m-%Y %H:%M:%S] "))
            handle.write(log_stamp + message.rstrip() + "\n")
            handle.close()

    def run_command(self, command):
        self.log("Running: %s" % command.strip() + "\n")
        output, errors = run_command(command)

        for line in errors:
            u_line = line.decode('utf-8')
            self.log(u"STDERR: " + u_line + "\n")

        for line in output:
            u_line = line.decode('utf-8')
            self.log(u"STDOUT: " + u_line + "\n")

        return output, errors

    def prepare_arguments(self, row):
        commit_settings = ""
        message = ""
        columns = self._model.get_columns()

        for field in ACTOR_FIELDS:
            if "commit" in field and not self._force_committed_date:
                # We're letting git update the committer's name and email
                continue

            index = Index(row=row, column=columns.index(field))
            value = self._model.data(index)
            commit_settings = add_assign(commit_settings, field, value)

        for field in TIME_FIELDS:
            if "commit" in field and not self._force_committed_date:
                # We're letting git update the committed date.
                continue

            index = Index(row=row, column=columns.index(field))
            _timestamp, _tz = self._model.data(index)
            value = str(_timestamp) + " " + _tz.tzname(None)
            commit_settings = add_assign(commit_settings, field, value)

        field = "message"
        index = Index(row=row, column=columns.index(field))
        message = self._model.data(index)

        return commit_settings, message

    def run(self):
        """
            Main method of the script. Launches the git command and
            logs if the option is set.
        """
        os.chdir(self._directory)
        try:
            self.pick_and_commit()
        except Exception:
            raise
        finally:
            self.cleanup_repo()
        self._finished = True

    def cleanup_repo(self):
        # Do some verifications before these cleanup steps.
        a_repo = Repo(self._directory)
        if a_repo.is_dirty():
            self.run_command('git reset HEAD --hard')

        try:
            branches = [branch.name for branch in a_repo.branches]

            if 'gitbuster_rebase' in branches:
                self.run_command('git checkout %s' % self._fallback_branch_name)
                self.run_command('git branch -D gitbuster_rebase')
        except TypeError:
            self.run_command('git checkout %s' % self._fallback_branch_name)

    def ref_update(self, commit):
        """
            Update the commit, probably since one of it's parents has changed.
        """
        model = self._model

        if model.is_deleted(commit):
            # If the commit has been deleted, skip it
            return True

        parents = model.c_data(commit, "parents")

        if len(parents) != 1:
            # This is a merge
            for parent in parents:
                if parent in self._should_be_updated and \
                   parent not in self._updated_refs:
                    # Meaning one of the parent branches of the merge hasn't
                    # been rewritten yet => skip for now
                    return True

        # The following will be useful to query the model
        commit_row = model.row_of(commit)

        # Here the parent should be the one defined in the mode (we should call
        # data()). Otherwise, we won't be able to insert or delete commits.
        _parent = parents[0]
        if _parent in self._updated_refs:
            _parent_sha = self._updated_refs[_parent]
        else:
            _parent_sha = self._model.c_data(_parent, "hexsha")

        self.run_command("git checkout -f %s" % _parent_sha)

        to_pick_hexsha = model.c_data(commit, "hexsha")
        if len(parents) == 1:
            # This is not a merge
            pick_command = "git cherry-pick -n %s" % to_pick_hexsha
        else:
            # This is a merge
            pick_command = "git cherry-pick -n -m 1 %s" % to_pick_hexsha

        output, errors = self.run_command(pick_command)
        if [line for line in errors if "error: could not apply" in line]:
            # We have a merge conflict.
            model.set_conflicting_commit(commit_row)
            commit = model.get_conflicting_commit()
            if commit in self._solutions:
                apply_solutions(self._solutions[commit])
            else:
                # Find out what were the hexsha of the conflicting commit
                # and of it's parent, in order to find the diff.
                self.process_unmerged_state(_parent_sha)
                self.cleanup_repo()
                return False

        output, errors = self.run_command("git write-tree")
        new_tree = output[0].strip()

        parent_string = ""
        for parent in parents:
            if parent in self._updated_refs:
                _parent = self._updated_refs[parent]
            else:
                _parent = self._model.c_data(parent, "hexsha")

            parent_string += "-p %s " % _parent

        FIELDS, MESSAGE = self.prepare_arguments(commit_row)
        with open("tmp_message", "w") as handle:
            handle.write(MESSAGE)

        output, errors = self.run_command(FIELDS +
                                        "git commit-tree %s %s < tmp_message" %
                                        (new_tree, parent_string))
        new_sha = output[0].strip()
        self._last_updated_sha = new_sha
        self._updated_refs[commit] = new_sha

        self._progress += 1. / self._to_rewrite_count
        for _commit in self._model.c_data(commit, "children"):
            if not self.ref_update(_commit):
                return False

        return True

    def pick_and_commit(self):
        """
            This is the method that actually does the rebasing.
        """
        self._should_be_updated = self.all_should_be_updated(self._start_commits)

        self._progress = 0
        for commit in self._start_commits:
            if not self.ref_update(commit):
                # There is a conflict
                return False

        # Update other references (as branches, tags)
        command = "git update-ref refs/heads/gitbuster_rebase %s"
        self.run_command(command % self._last_updated_sha)
        output, errors = self.run_command("git checkout gitbuster_rebase")

        if errors:
            # Something, somewhere, went very wrong.
            error = "Something went wrong in the git filter/rebase process, " +\
                    "not applying your changes to avoid data loss, please " +\
                    "check the logs."
            raise GfbiException(error)

        if self._model.is_name_modified():
            # The model may be fake
            new_branch_name = self._model.get_new_branch_name()
            self.run_command('git branch -M %s' % new_branch_name)
            if not self._model.is_fake_model():
                self.run_command('git branch -D %s' % self._branch.name)

            branches = Repo(self._directory).branches
            new_branch = [branch for branch in branches
                          if branch.name == new_branch_name][0]
            self._model.set_current_branch(new_branch, force=True)

            # Setting the _branch to the new branch
            # This is useful if we create a fake branch, and change it's name
            # after the creation.
            self._branch = new_branch
        else:
            self.run_command('git branch -M %s' % self._branch.name)

        if not self._dont_populate:
            self._model.populate()
        self._success = True

    def process_unmerged_state(self, orig_hexsha):
        """
            Process the current unmerged state and inform the model about
            unmerged files.
        """
        model = self._model
        conflicting_commit = model.get_conflicting_commit()
        hexsha = model.c_data(conflicting_commit, "hexsha")

        self._u_files = get_unmerged_files(hexsha, orig_hexsha, self._directory)
        self._model.set_unmerged_files(self._u_files)

    def progress(self):
        """
            Returns the progress percentage
        """
        return self._progress

    def output(self):
        """
            Returns the output as a list of lines
        """
        return list(self._output)

    def errors(self):
        """
            Returns the errors as a list of lines
        """
        return list(self._errors)

    def is_finished(self):
        """
            Returns self._finished
        """
        return self._finished

    def is_success(self):
        """
            Returns True if the process went through without failing.
        """
        return self._success



from os import waitpid
from subprocess import Popen

def validate_branch_name(candidate):
    """
    Checks whether the name change is legitimate


    >>> validate_branch_name("master master")
    Traceback (most recent call last):
    ...
    ValueError: Invalid name: has spaces in it
    >>> validate_branch_name("")
    Traceback (most recent call last):
    ...
    ValueError: Invalid name: got empty string
    >>> validate_branch_name("azertyop")
    >>> validate_branch_name("azertyop/qsdfghjklm/wxcvb")
    >>> validate_branch_name("azertyop/azertop//")
    Traceback (most recent call last):
    ...
    ValueError: Invalid name, \
see http://www.kernel.org/pub/software/scm/git/\
docs/git-check-ref-format.html
    >>> validate_branch_name("azertyop/.qsfglm")
    Traceback (most recent call last):
    ...
    ValueError: Invalid name, \
see http://www.kernel.org/pub/software/scm/git/\
docs/git-check-ref-format.html
    """
    if not candidate:
        raise ValueError("Invalid name: got empty string")
    if any (item in candidate for item in " \t\n"):
        raise ValueError("Invalid name: has spaces in it")
    command = u'git check-ref-format refs/tags/%s' % candidate
    process = Popen(command.split())
    status = waitpid(process.pid, 0)[1]
    if status != 0:
        raise ValueError("Invalid name, see http://www.kernel.org/"
            "pub/software/scm/git/docs/git-check-ref-format.html")
